import asyncio
import hashlib
import secrets

from aiohttp import web

import config
import web_interface.page_former
import user_login
import version
import http_client


SHOW_TOKENS_IN_API = False


def find_printer_interface(app, printer_id):
    for pi in app.printer_interfaces:
        if not printer_id or pi.id_string == printer_id:
            return pi


def printer_report(pi):
    report = {}
    report['report'] = pi.status_report()
    report['errors'] = pi.get_errors_to_send()
    http_client.remove_repeted_errors(report)
    reg_code = pi.registration_code
    if reg_code:
        report['reg_code'] = reg_code
    if SHOW_TOKENS_IN_API:
        report['token'] = pi.printer_token
    return report


def get_static_values(app):
    values = {'version' : version.full_version_string()}
    values['cloud_url'] = config.get_settings()['URL'].strip("cli-")
    values['style_name'] = config.get_settings().get("web_interface", {}).get("style", "style")
    values['show_your_account_button'] = not app.offline_mode
    return values


def full_report(app):
    response_dict = {'printers': {}}
    for pi in app.printer_interfaces:
        response_dict['printers'][pi.id_string] = printer_report(pi)
        if getattr(app, "updater"):
            response_dict['update_available'] = app.updater.update_available
            response_dict['update_downloaded'] = app.updater.update_downloaded
    return response_dict


def printer_command(path_list, args, app):
    error = ''
    body = ''
    status = web.HTTPOk.status_code
    if len(path_list) > 1:
        printer_id = path_list[1]
    else:
        printer_id = args.get('printer_id')
    if not printer_id:
        try:
            pi = app.printer_interfaces[0]
        except IndexError:
            pi = None
    else:
        pi = find_printer_interface(app, printer_id)
        if not pi:
            error == 'No printer with id ' + str(printer_id)
            status = web.HTTPClientError.status_code
    if pi:
        if 'command' in args or path_list[2] == 'command':
            if len(path_list) > 2:
                command = path_list[3]
            else:
                command = args.get('command')
            if command:
                command_args = {'command': command, 'number': -1}
                payload = {}
                payload.update(args)
                if 'command' in payload:
                    del payload['command']
                if 'printer_id' in payload:
                    del payload['printer_id']
                if payload:
                    command_args['payload'] = payload
                ack = pi.execute_server_command(command_args)
                if not ack['result']:
                    body = {'error': {'message': 'Command failure',
                                      'code': web.HTTPServerError.status_code,
                                      'ack' : ack}}
                else:
                    body = ack
            else:
                status = web.HTTPClientError.status_code
                error = 'No command'
        elif len(path_list) < 3 or 'report' in path_list or 'report' in args or path_list[2] == '':
            body = printer_report(pi)
            if not body:
                status = web.HTTPServerError.status_code
                error = 'The printer has disconnected'
        else:
            status = web.HTTPBadRequest.status_code
            error = 'No such command'
    return error, body, status


def process_login(args, app):
    login = args.get('login', '').replace(' ', '+') # a hack to fix + in email.
    password = args.get('password', '')
    password_hash = args.get('password_hash')
    print("!!!args", args)
    if not password_hash:
        password_hash = hashlib.sha256(password.encode('utf-8')).hexdigest()
    auth_token = args.get('auth_token')
    if not login and not auth_token:
        return "No login"
    correct_login, correct_password_hash = user_login.UserLogin.load_login()
    if correct_login:
        if login != correct_login or password_hash != correct_password_hash:
            return "Incorrect credentials"
    elif not app.user_login.got_network_connection:
        return "No network to check login credentials on the server"
    else:
        error = app.user_login.login_as_user(login, password_hash, auth_token)
        if error:
            return error[1]


async def api2_handler(request):
    body = {}
    path_list = request.path.split('/')[2:]
    status = web.HTTPOk.status_code
    args = {}
    args.update(request.query)
    app = request.app['3dp_app']
    if request.method == 'POST':
        if request.content_type == 'application/json':
            args.update(await request.json())
        else:
            args.update(await request.post())
    print("!", path_list, args)
    if path_list and path_list[0] == 'login':
        error = process_login(args, request.app['3dp_app'])
        if error:
            status = web.HTTPNonAuthoritativeInformation.status_code
            await asyncio.sleep(web_interface.handlers.AUTH_ERROR_SLEEP)
        else:
            new_cookie_token = secrets.token_hex(web_interface.page_former.PageFormer.COOKIE_TOKEN_LENGHT)
            request.app['cookie_tokens'].append(new_cookie_token)
            cookie_header_value = f"token:{new_cookie_token}; HttpOnly"
            if request.secure:
                cookie_header_value += '; Secure'
            body = {'auth_token': new_cookie_token}
            return web.json_response(status=status,
                                 data=body,
                                 headers={'Set-cookie': cookie_header_value})
    else:
        error = web_interface.handlers.check_auth(request)
        provided_auth_token = args.get('auth_token')
        if not error or (provided_auth_token and provided_auth_token in request.app['cookie_tokens']):
            if 'full_report' in path_list:
                body = full_report(app)
            elif 'static_values' in path_list:
                body = get_static_values(app)
            elif 'settings' in path_list:
                if request.method == "GET":
                    body = config.get_settings()
                elif request.method == "POST":
                    settings = config.get_settings()
                    try:
                        config.merge_dictionaries(settings, args)
                        config.Config.instance().save_settings(settings)
                    except Exception as e:
                        error = "Error on saving settings"
                        status = web.HTTPServerError
            elif 'printer_id' in args or 'printer_id' in path_list:
                error, body, status = printer_command(path_list, args, app)
            elif 'logout' in path_list:
                app.user_login.logout()
                app.restart_flag = True
                app.stop_flag = True
        else:
            status = web.HTTPUnauthorized.status_code
    if error:
        body = {'error': {'message': error, 'code': status}}
    return web.json_response(status=status, data=body)
